#define INIT_DUTY_CYCLE 0   // Ciclo de trabajo inicial
#define PRESC 0
#define DIVA 42
#define CPRD 2000
#define N_SAVED 3000

#define N_Tpwm 11
#define delta_Ref	0.002
#define REF50 0.5f
#define REF75 0.75f
#define REF80 0.8f
#define REF100 1.f
#define PIO_LED PIOC
#define LED_1 PIO_PC25
#define LED_2 PIO_PC24 
#define PIO_BOTON PIOC
#define B_START PIO_PC23
#define B_STOP PIO_PC22

#include <Arduino.h>

//Declaración de funciones
void GPIO_setup(void);
void ADC_setup(void);
void PWM_setup(void);
void PWM_update_duty_cycle(float duty_cycle);
void ADC_Handler(void);


// Variables globales
uint16_t  adc_7;
bool run, start_saving, data_saved, saving;
float s_ref[N_SAVED], s_y[N_SAVED], s_u[N_SAVED];

void setup() {
  // Llamada a funciones de configuración
  GPIO_setup();
  ADC_setup();
  PWM_setup();
  Serial.begin(115200);
}

void loop() {

  static bool bk, bkm1 = true, b2k, b2km1 = true;
  bk = PIO_Get(PIO_BOTON,PIO_INPUT,B_START);
  b2k = PIO_Get(PIO_BOTON,PIO_INPUT,B_STOP);
  if(bk == 1 && bkm1 == 0){
    // Se produjo una pulsación del botón de arranque
    run = 1;
    if(!saving){
      start_saving = 1;
    }
    

  }
  if(b2k == 1 && b2km1 == 0){
    // Se produjo una pulsación del botón de parada
    run = 0;

  }
  b2km1 = b2k;
  bkm1 = bk;
  if(data_saved){
    data_saved = 0;
    Serial.print("start");
    Serial.print('\n');
    for(uint16_t i = 0 ; i < N_SAVED ; i++){
      Serial.print(s_ref[i],4);
      Serial.print(',');
      Serial.print(s_y[i],4);
      Serial.print(',');
      Serial.print(s_u[i],4);
      Serial.print('\n');
      delay(10);
    }
    Serial.print("end");
    Serial.print('\n');
  }

}

void GPIO_setup(void){
  PMC->PMC_PCER0 |= PMC_PCER0_PID13;
  // Configuración de pin de salida
  // Se configuran los pines PC24 (D6) y PC25 (D5) como salidas
  PIO_SetOutput(PIO_LED,LED_1|LED_2,0,0,0);


  // Configuración de pin de entrada
  // Se configura primero el debouncing en el pin PC23 (D7)
  
  PIO_SetDebounceFilter(PIO_BOTON,B_START,10); //Se usa una frecuencia de filtro de 50 Hz. Se puede aumentar o disminuir para mejorar el debounce
  PIO_SetInput(PIO_BOTON,B_START,PIO_DEBOUNCE);
  PIO_SetDebounceFilter(PIO_BOTON,B_STOP,10); //Se usa una frecuencia de filtro de 50 Hz. Se puede aumentar o disminuir para mejorar el debounce
  PIO_SetInput(PIO_BOTON,B_STOP,PIO_DEBOUNCE);

}



void ADC_setup(void){
  // Configuración del ADC
  // Primero se activa desde el control de energía (PMC)
  PMC->PMC_PCER1 |= PMC_PCER1_PID37;
  // Reset del ADC para limpiar cualquier valor espurio
  ADC->ADC_CR = ADC_CR_SWRST;

/*
  En el siguiente registro se configuran los parámetros más importantes del ADC
  PRESCAL: Prescaler del reloj. El reloj máximo del ADC no debe superar 22 MHz. Por defecto el prescaler mínimo es 2.
    ADCClock = MCK/((PRESCAL + 1)*2). Para MCK = 84 MHz con PRESCAL = 1 , ADCClock = 21 MHz
    t_conv = 20/ADCClock = 0.95238 ns -> 1 us mínimo para 1 Msps
  TRGEN y TRGSEL_ADC_TRIG4: Se selecciona como trigger una fuente de hardware, en este caso la línea de evento del PWM
  USEQ_REG_ORDER: Se selecciona el muestreo secuencial
  TRACKTIM: Tiempo de "tracking". Es el tiempo entre el inicio de conversión de cada canal. 
    Se selecciona un valor mayor al de conversión para asegurar la conversión correcta a 12 bits.
    Tracking = (TRACKTIM + 1)/ADCClock -> con TRACTIM = 30 se obtiene 1,43 us
  TRANSFER: Tiempo de "transferencia". Es el tiempo entre el trigger de conversión del ADC en un nuevo canal, y el inicio actual de la conversión (ver datasheet).
    Se selecciona un valor pequeño, para asegurar una correcta transferencia de canal a canal.
    Transfer = (TRANSFER*2+3)/ADCClock -> con TRANSFER = 3 se obtienen 0,43 us
  Con el tracking y la transferencia, el tiempo de conversión total por canal es de aprox. 2 us.
  Como son 2 canales, ref(t) y y(t), el tiempo de conversión total antes del cálculo de la uc es de 4 us.
  Con un periodo de 1 ms, se corresponde a un 0,4% del periodo.
  Las otras configuraciones se dejan por defecto
*/ 

  ADC->ADC_MR |= ADC_MR_PRESCAL(1) | ADC_MR_TRGEN | ADC_MR_TRGSEL_ADC_TRIG4 | ADC_MR_TRACKTIM(30) | ADC_MR_TRANSFER(3);// | ADC_MR_USEQ_REG_ORDER; No se usan los 2 canales

  // Se activan los canales a muestrear
  ADC->ADC_CHER = ADC_CHER_CH7;  
  //ADC->ADC_CHER = ADC_CHER_CH6 | ADC_CHER_CH7;
  // Se selecciona la secuencia de muestreo. En este caso se usan los canales 6 y 7 para concordar con A0 y A1 en la placa
  //ADC->ADC_SEQR1 = ADC_SEQR1_USCH7(6) | ADC_SEQR1_USCH8(7); // Debe muestrear 6, luego 7 y finalizar
  //ADC->ADC_SEQR2 = 0;
  // Se configura la interrupción de fin de conversión del canal 7, para ese entonces el canal 6 también debe haber finalizado
  ADC->ADC_IER |= ADC_IER_EOC7;
  NVIC_EnableIRQ(ADC_IRQn);

}

void ADC_Handler(void){
  static float ref = 0; // Valor de referencia
  static float yk = 0;    // Valor de la salida actual y(k)
  static float uk = 0;    // Valor de acción de control
  /*-------------------DEFINIR VARIABLES NECESARIAS--------------------------------*/

  
  static uint16_t index = 0, counter_N = 0, counter_ref1 = 0;


  counter_N++; //Cada vez que se ejecuta la interrupción del ADC se aumenta el contador para obtener la frecuencia de muestreo a partir de la del PWM
  
  
  // Primero se verifica si la bandera de fin de conversión corresponde al canal 7 del ADC
  // Y si corresponde ejecutar el cálculo de la acción de control según N_Tpwm
  if(ADC->ADC_ISR & ADC_ISR_EOC7){  
    adc_7 = ADC->ADC_CDR[7]; //Se lee el valor del registro de resultado de conversión del canal 7
    if(counter_N == N_Tpwm){
      PWM_update_duty_cycle(uk); // Se actualiza la acción de control al inicio del periodo de muestreo para
                                // simular el atraso de implementación digital de un periodo completo

      PIO_Set(PIO_LED,LED_1); // Se enciende el LED de estado 1
    
      //El siguiente código se ejecuta solamente si se dio la señal de arranque del motor
      if(run){

        // Variación de referencia en rampa ascendente
        // ---------------------------------------------------------------------------------------
        // --------------------------------ATENCION!!!!!-----------------------------------
        // Esto debe modificarse para obtener la variación de referencia solicitada
        // en la guía de laboratorio.
        counter_ref1++;
        if(ref < REF50){ 
          ref = ref + delta_Ref;
        }else{ 
          ref = REF50;								 
        }
        // ESCALÓN de referencia
        if(counter_ref1 >= 500){	 	// Si el contador alcanzó t = 500*N_Tpwm*0.001s, la ref.
        ref = REF80; // sube del 50% al 75%
        }

        //------------ Normalización del valor del ADC y cálculo del error ------------                 
        yk = ((float) adc_7)/3276.f;


        //------------ Cálculo de la acción de control --------------------------------




        // Aplica acción de control
        // ---------------------------------------------------------------------------------------
        // --------------------------------ATENCION!!!!!-----------------------------------
        // Comentar la acción de control que NO se desea aplicar
        //uk = up_k;		// P
        //uk = upi_k;		// PI
        //uk = upid_k;		// PI-D
        //uk = udb_k;		// deadbeat
        uk = 1; //Lazo abierto
        if(uk<0){
          uk = 0;
        }
        if(uk>1){
          uk = 1;
        }
        //PWM_update_duty_cycle(uk);
        PIO_Clear(PIO_LED,LED_1); //El led se apaga al actualizar la acción de control

        //-------- Valores anteriores para el proximo periodo de muestreo-----------------


        //-------------------------------------------------------------------------------
        
        if(start_saving && !saving){
          // Si se da la bandera de iniciar el guardado de datos, se limpia dicha bandera,
          // se activa la bandera de guardado en progreso, y se inicializa el indice para el vector de datos en 0
          start_saving = 0;
          saving = 1;
          index = 0;
        }
        // Si se dio la señal de guardado de variables, se guardan utilizando el indice hasta que los arreglos estén llenos
        if(saving){
          PIO_Set(PIO_LED,LED_2);
          
          s_ref[index] = ref;
          s_y[index] = yk;
          s_u[index] = uk;
          
          /*
            Modificar las variables guardadas para guardar acciones de control PI y D por separado.
          */
          index++;
          if(index>=N_SAVED){
            saving = 0;
            data_saved = 1;
          }
        } else {
          PIO_Clear(PIO_LED,LED_2);
        }

      } else {
        /*-------------------------Limpiar variables creadas para el cálculo de acción de control------------*/


        /*---------------------------------------------------------------------------------------------------*/
        // Si el motor no arrancó, o se presionó el boton de parada, se limpian las variables involucradas

        uk = 0;
        index = 0;
        if(saving){
        saving = 0;
        data_saved = 1;
        }
        ref = 0.f;
        counter_ref1 = 0;
        PWM_update_duty_cycle(0);
        PIO_Clear(PIO_LED,LED_2);
      }

    counter_N = 0; // Se limpia el contador que genera el periodo de muestreo
    }
  }

}



void PWM_setup(void){
  // Primero se debe activar la macrocelda PWM desde el controlador de energía (PMC). En este caso el módulo PWM tiene el ID 36
  PMC->PMC_PCER1 |= PMC_PCER1_PID36;
  // Luego, se configura el periférico del canal 0 en el pin correspondiente B16
  PIOB->PIO_ABSR |= PIO_ABSR_P16;
  // Se le da control del pin al periférico
  PIOB->PIO_PDR |= PIO_PDR_P16;
  {
      // Luego, se configura el periférico del canal 0 en el pin correspondiente B12
  PIOB->PIO_ABSR |= PIO_ABSR_P12;
  // Se le da control del pin al periférico
  PIOB->PIO_PDR |= PIO_PDR_P12;

  }
  //------------- Configuración específica del timer
  // Primero configura el reloj. El módulo PWM puede generar 2 relojes subdividiendo el del del sistema, clkA y clkB.
  // En este caso, se usa solo clkA
  // Reloj del sistema MCK = 84 MHz. Para PWM con alineación izquierda F_PWM = MCK/(CRPD*PRESC*DIVA) donde CRPD es el "periodo" o cantidad de conteos
  // Por lo tanto, se configura primero PRESC y DIVA 
  PWM->PWM_CLK = PWM_CLK_PREA(PRESC)|PWM_CLK_DIVA(DIVA); // Con esta configuración, para F_PWM de 1 kHz, el contador CRPD llega a 2000, es decir, resolución de 0.05% en ciclo de trabajo
  // Una vez configurado el reloj, se selecciona el canal PWM, la alineación y la polaridad. En este caso, el canal 0, polaridad es por defecto alineación izquierda
  PWM->PWM_CH_NUM[0].PWM_CMR = PWM_CMR_CPRE_CLKA | PWM_CMR_CPOL;
  // Se carga el valor máximo del contador, CRPD, correspondiente al "periodo"
  PWM->PWM_CH_NUM[0].PWM_CPRD = CPRD;  // F_PWM = 84000000/(2000*1*42) = 1 kHz
  // Se configura un valor de ciclo útil inicial y activa el canal del PWM
  PWM->PWM_CH_NUM[0].PWM_CDTY= (float)INIT_DUTY_CYCLE/100.0f*CPRD;

  // La sig. config. se corresponde al evento para iniciar el muestreo del ADC
  // El evento no se activa con una bandera de desbordamiento, sino con comparación directa con el valor del contador. 
  // Para que ocurra cada inicio de periodo, se compara con el valor máximo del contador, en este caso CPRD
  PWM->PWM_CMP[0].PWM_CMPV = PWM_CMPV_CV(CPRD);
  // Se activa la comparación
  PWM->PWM_CMP[0].PWM_CMPM = PWM_CMPM_CEN;      // Enable comparison
  
  // Al final,  se selecciona el canal 0 en el generador de pulsos de la línea de eventos 0 (se puede usar otra línea de eventos)
  PWM->PWM_ELMR[0] = PWM_ELMR_CSEL0; 

  // Se activa el PWM
  PWM->PWM_ENA = PWM_ENA_CHID0;   
}



/**
 * \brief Modifica el registro de actualización de ciclo útil. El cambio tendrá efecto en el siguiente periodo.
 *
 * \param duty_cycle              Valor de ciclo útil, en unidad.
 */
void PWM_update_duty_cycle(float duty_cycle){
  // En este dispositivo, el cambio del ciclo útil se realiza sólamente al inicio de cada ciclo
  // Es decir, siempre existe un atraso de implementación digital de un periodo
  if (duty_cycle >= 1){
    duty_cycle = 1;
  }
  PWM->PWM_CH_NUM[0].PWM_CDTYUPD = duty_cycle*CPRD;
}


